#include <AppKit/AppKit.h>
#include "AddressView/ADPersonView.h"
#include "Controller.h"
#include "DragDropMatrix.h"

@interface Controller (Private)
- (void) browserAction: (id) sender;
@end

@implementation Controller
- (void) applicationDidFinishLaunching: (NSNotification*) note
{
  [NSApp registerServicesMenuSendTypes: [NSArray arrayWithObjects:
						   NSStringPboardType,
						 nil]
	 returnTypes: nil];
  
  servicesMenu = [[[NSApp mainMenu]
		    itemWithTitle: @"Services"]
		   submenu];
  [NSApp setServicesMenu: servicesMenu];

  [groupsBrowser selectRow: 0 inColumn: 0];
  [groupsBrowser selectRow: 0 inColumn: 1];

  [[NSNotificationCenter defaultCenter]
    addObserver: self
    selector: @selector(handleDatabaseChanged:)
    name: ADDatabaseChangedNotification
    object: nil];
  [[NSNotificationCenter defaultCenter]
    addObserver: self
    selector: @selector(handlePersonDroppedOnGroup:)
    name: DDMOperationConcludedNotification
    object: nil];
}

- (void) awakeFromNib
{
  _fm = [NSFileManager defaultManager];
  _book = [ADAddressBook sharedAddressBook];

  [groupsBrowser setAllowsEmptySelection: NO];
  [groupsBrowser setAllowsMultipleSelection: NO];
  [groupsBrowser setMaxVisibleColumns: 2];
  [groupsBrowser setDelegate: self];
  [groupsBrowser setTitle: _(@"Group") ofColumn: 0];
  [groupsBrowser setTitle: _(@"Name") ofColumn: 1];
  [groupsBrowser setMatrixClass: [DragDropMatrix class]];
  [groupsBrowser setTarget: self];
  [groupsBrowser setAction: @selector(browserAction:)];

  clipView = [[NSClipView alloc] initWithFrame: [scrollView frame]];
  [clipView setAutoresizesSubviews: YES];
  personView = [[ADPersonView alloc] initWithFrame: NSZeroRect];
  [clipView setDocumentView: personView];
  [personView setFillsSuperview: YES];
  [scrollView setContentView: clipView];
  [scrollView setHasVerticalScroller: YES];
  [scrollView setHasHorizontalScroller: YES];
  [scrollView setBorderType: NSBezelBorder];

  [self createCache];
  if([_peopleCache count])
    [self selectPerson: [_peopleCache objectAtIndex: 0]];
}

- (void) createCache
{
  [_peopleCache release];
  if(_currentGroup)
    _peopleCache = [_currentGroup members];
  else
    _peopleCache = [_book people];
  _peopleCache =
    [[_peopleCache sortedArrayUsingSelector: @selector(compareByScreenName:)]
      retain];
}

- (NSArray*) groupNames
{
  NSArray *groups = [_book groups];
  NSMutableArray *retval = [NSMutableArray arrayWithCapacity: [groups count]];

  int i;
  for(i=0; i<[groups count]; i++)
    [retval addObject: [[groups objectAtIndex: i]
			 valueForProperty: ADGroupNameProperty]];
  return retval;
}

- (void) selectPerson: (ADPerson*) person
{
  int i;

  if(!person)
    {
      [personView setPerson: nil];
      return;
    }
  
  for(i=0; i<[_peopleCache count]; i++)
    if([[[_peopleCache objectAtIndex: i] uniqueId]
	 isEqualToString: [person uniqueId]])
      break;
  
  if(i==[_peopleCache count]) // Not found? Select "all" and try again
    {
      [groupsBrowser selectRow: 0 inColumn: 0]; 
      for(i=0; i<[_peopleCache count]; i++)
	if([[[_peopleCache objectAtIndex: i] uniqueId]
	     isEqualToString: [person uniqueId]])
	  break;
    }

  if(i==[_peopleCache count]) // Still not found? WEIRD.
    {
      NSLog(@"Person %@ not found!\n", [person uniqueId]);
      return;
    }

  [groupsBrowser selectRow: i inColumn: 1];
  [personView setPerson: [_peopleCache objectAtIndex: i]];
  [clipView scrollToPoint: NSMakePoint(0, [personView frame].size.height)];
}

- (void) deletePersonAndSelectNext: (ADPerson*) person
{
  int row = [groupsBrowser selectedRowInColumn: 1];

  if(_currentGroup)
    {
      NSLog(@"Deleting from group\n");
      [_currentGroup removeMember: person];
    }
  else
    {
      NSLog(@"Deleting from all\n");
      [_book removeRecord: person];
    }
  [self createCache];
  [groupsBrowser reloadColumn: 1];

  if(![_peopleCache count])
    [personView setPerson: nil];
  else if(row >= [_peopleCache count])
    [self selectPerson: [_peopleCache objectAtIndex: [_peopleCache count]-1]];
  else
    [self selectPerson: [_peopleCache objectAtIndex: row]];
}

- (void) beginEditingPerson: (ADPerson*) person
{
  [self selectPerson: person];
  [personView setEditable: YES];
  [editItem setTitle: _(@"End editing")];
  [editButton setState: NSOnState];
  [clipView scrollToPoint: NSMakePoint(0, [personView frame].size.height)];
};

- (void) finishEditingPerson
{
  ADPerson *p = [personView person];
  if(!p || ![personView isEditable]) return;

  [clipView scrollToPoint: NSMakePoint(0, [personView frame].size.height)];

  if(![p valueForProperty: ADLastNameProperty] ||
     ![p valueForProperty: ADLastNameProperty])
    {
      if(NSRunAlertPanel(_(@"Discard Person?"),
			 _(@"The person you have edited has no first or last\n"
			   @"names. Would you like to discard this person?"),
			 _(@"Yes"), _(@"No"), nil, nil))
	[self deletePersonAndSelectNext: p];
    }
  [personView setEditable: NO];
  [editItem setTitle: _(@"Edit Person")];
  [editButton setState: NSOffState];

  [groupsBrowser reloadColumn: 1];
  [self selectPerson: p];
}

/*
  Actions
*/

- (void) doEditPerson: (id) sender
{
  ADPerson *p = [personView person];
  if(!p) return; // nothing to do
  if([personView isEditable])
    [self finishEditingPerson];
  else
    [self beginEditingPerson: p];
  [clipView scrollToPoint: NSMakePoint(0, [personView frame].size.height)];
}

- (void) doTogglePersonEditable: (id) sender
{
  ADPerson *p = [personView person];
  if(!p) return; // nothing to do
  if([sender state] == NSOnState)
    [self beginEditingPerson: p];
  else
    [self finishEditingPerson];
}

- (void) doCreatePerson: (id) sender
{
  // select "All"
  [groupsBrowser selectRow: 0 inColumn: 0];
  [self browserAction: groupsBrowser];

  ADPerson *p = [[ADPerson alloc] init];
  [_book addRecord: p];
  [self createCache];
  [groupsBrowser reloadColumn: 1];
  [self beginEditingPerson: p];
}

- (void) doDeletePerson: (id) sender
{
  ADPerson *p = [personView person];
  if(!p) return; // nothing to do.

  NSString *msg;
  if(_currentGroup)
    msg = [NSString stringWithFormat: _(@"Do you really want to delete %@ "
					@"from %@?\n"
					@"(%@ will remain in \"All\".)"),
		    [p screenName],
		    [_currentGroup valueForProperty: ADGroupNameProperty],
		    [p screenName]];
  else
    msg = [NSString stringWithFormat: _(@"Do you really want to delete %@ "
					@"from \"All\" and all groups?"),
		    [p screenName]];
  if(!NSRunAlertPanel(_(@"Delete Person?"), msg,
		      _(@"Yes"), _(@"No"), nil, nil))
    NSLog(@"Not deleting.\n");
  else
    [self deletePersonAndSelectNext: p];
}

- (void) doImportPerson: (id) sender
{
  ADConverterManager *man = [ADConverterManager sharedManager];
  
  NSOpenPanel *p = [NSOpenPanel openPanel];
  [p setCanChooseDirectories: NO];
  [p setAllowsMultipleSelection: NO];
  int retval = [p runModalForTypes: [man inputConvertableFileTypes]];
  if(!retval)
    return;

  NSString *fname = [[p filenames] objectAtIndex: 0];
  id<ADInputConverting> conv = [man inputConverterWithFile: fname];
  NSAssert(conv, @"No converter for this file!");

  id obj;
  int loaded = 0;

  while((obj = [conv nextRecord]))
    {
      loaded++;
      
      NSEnumerator *e = [_peopleCache objectEnumerator];
      id other;
      int retval;

      retval = 0; // insert anyway
      while((other = [e nextObject]))
	{
	  if([[other screenName] isEqualToString: [obj screenName]])
	    {
	      NSString *fmt; 
	      fmt =
		[NSString stringWithFormat:
			    _(@"Trying to import person named '%@',\n"
			      @"which already exists in the database."),
			  [obj screenName]];
	      retval = NSRunAlertPanel(_(@"Existing person?"),
				       fmt, _(@"Replace"), _(@"Insert anyway"),
				       _(@"Don't insert"), nil);
	      break;
	    }
	}
      if(retval == 1) // replace
	{
	  [_book removeRecord: other];
	  [_book addRecord: obj];
	}
      else if(retval == 0) // insert anyway
	[_book addRecord: obj];
      else if(retval == -1) // don't insert; continue reading
	continue;

    }
  [groupsBrowser reloadColumn: 1];
}

- (void) doCreateGroup: (id)sender
{
  ADGroup *group = [[[ADGroup alloc] init] autorelease];
  [group setValue: _(@"New Group") forProperty: ADGroupNameProperty];
  if(![_book addRecord: group])
    NSRunAlertPanel(_(@"Error"), _(@"Could not create group!"),
		    _(@"OK"), nil, nil, nil);
  [groupsBrowser reloadColumn: 0];
}

- (void) doDeleteGroup: (id) sender
{
  int row = [groupsBrowser selectedRowInColumn: 0];
  if(row == 0) return; // Can't delete this
  ADGroup *group = [[_book groups] objectAtIndex: row-1];
  NSString *name = [group valueForProperty: ADGroupNameProperty];

  NSString *msg =
    [NSString stringWithFormat: _(@"Do you really want to delete "
				  @"the group '%@'?"),
	      name];
  int retval = NSRunAlertPanel(_(@"Delete Group?"), msg,
			       _(@"Yes"), _(@"No"), nil, nil);
  if(!retval) return;

  if(![_book removeRecord: group])
    {
      msg = 
	[NSString stringWithFormat: _(@"The group '%@` could not be deleted."),
		  name];
      NSRunAlertPanel(_(@"Error"), msg, _(@"OK"), nil, nil, nil);
    }

  lastCell = nil;
  [groupsBrowser reloadColumn: 0];
  [groupsBrowser selectRow: row-1 inColumn: 0]; // OK -- we caught 0 above
  [self browserAction: groupsBrowser];
  [self createCache];
  if([_peopleCache count])
    [self selectPerson: [_peopleCache objectAtIndex: 0]];
}

- (void) doSaveDatabase: (id) sender
{
  if([personView isEditable])
    [self finishEditingPerson];
  if([_book hasUnsavedChanges])
    {
      if(![_book save])
	NSRunAlertPanel(_(@"Couldn't save"),
			_(@"The database could not be saved!"),
			_(@"OK"), nil, nil, nil);
      else
	[[personView window] setTitle: _(@"Addresses")];
    }
}

/*
 *  Browser delegate methods
 */

- (int) browser: (NSBrowser*) sender
numberOfRowsInColumn: (int) column
{
  NSArray *groupnames = [self groupNames];

  if(column == 0)
    return [groupnames count]+1;
  else
    {
      NSCell *cell = [sender selectedCellInColumn: 0];
      NSString *oldName = @"";
      NSString *newName = @"";
      ADGroup *group = nil;

      if([sender selectedRowInColumn: 0] != 0)
	{
	  group =
	    [[_book groups] objectAtIndex: [sender selectedRowInColumn: 0]-1];
	  oldName = [group valueForProperty: ADGroupNameProperty];
	  newName = [cell stringValue];
	}
      
      // stop editing; rename if necessary
      [lastCell setEditable: NO];
      if(cell == lastCell && ![oldName isEqualToString: newName])
	{
	  if([newName isEqualToString: _(@"All")])
	    {
	      [cell setStringValue: oldName];
	      NSRunAlertPanel(_(@"Disallowed"),
			      _(@"You cannot rename a group to \"All\",\n"
				@"since that name is reserved by the system."),
			      _(@"OK"), nil, nil, nil);
	    }
	  else if([groupnames containsObject: newName])
	    {
	      [cell setStringValue: oldName];
	      NSRunAlertPanel(_(@"Disallowed"),
			      [NSString stringWithFormat:
					  _(@"You cannot rename this group "
					    @"to \"%@\",\n"
					    @"since a group of that name "
					    @"already exists."), newName],
			      _(@"OK"), nil, nil, nil);
	    }
	  else
	    [group setValue: newName forProperty: ADGroupNameProperty];
	  [lastCell setEditable: NO];
	}

      if(![newName isEqualToString: _(@"All")] &&
	 ![oldName isEqualToString: _(@"All")] &&
	 ![[cell stringValue] isEqualToString: _(@"All")])
	{
	  if([cell isEditable])
	    [cell setEditable: NO];
	  else
	    [cell setEditable: YES];

	  [lastCell release];
	  lastCell = [cell retain];
	}

      return [_peopleCache count];
    }

  return 0;
}

- (void) browser: (NSBrowser*) sender
 willDisplayCell: (id) cell
	   atRow: (int) row
	  column: (int) column
{
  if(column == 0)
    {
      if(row == 0)
	[cell setStringValue: _(@"All")];
      else
	{
	  [cell setStringValue: [[self groupNames] objectAtIndex: row-1]];
	  [cell setEditable: NO];
	}
    }
  else
    {
      [cell setStringValue: [[_peopleCache objectAtIndex: row] screenName]];
      [cell setLeaf: YES];
    }      
}

- (void) browserAction: (id) sender
{
  int row;

  if([personView isEditable])
    {
      [personView setPerson: nil];
      [personView setEditable: NO];
      [editButton setState: NSOffState];
    }
  
  if([sender selectedColumn] == 0)
    {
      row = [sender selectedRowInColumn: 0];
      ADGroup *group = nil;
      [_currentGroup release];
      _currentGroup = nil;
      
      if(row != 0)
	group = [[_book groups] objectAtIndex: row-1];
      if(![[group uniqueId] isEqualToString: [_currentGroup uniqueId]])
	{
	  _currentGroup = [group retain];
	  [self createCache];
	  [sender reloadColumn: 1];
	  if([_peopleCache count])
	    [self selectPerson: [_peopleCache objectAtIndex: 0]];
	  else
	    [self selectPerson: nil];
	}
    }
  else
    {
      row = [sender selectedRowInColumn: 1];
      ADPerson *person = [_peopleCache objectAtIndex: row];
      [personView setPerson: person];
    }
}

- (void) handleDatabaseChanged: (NSNotification*) note
{
  [[personView window] setDocumentEdited: YES];
  [[personView window] setTitle: _(@"Addresses*")];
  [self createCache];
}

- (void) handlePersonDroppedOnGroup: (NSNotification*) note
{
  NSDictionary *info = [note userInfo];
  int groupIndex = [[info objectForKey: @"DestinationRow"] intValue] - 1;
  int personIndex = [[info objectForKey: @"SourceRow"] intValue];

  ADPerson *p = [_peopleCache objectAtIndex: personIndex];
  ADGroup *g = [[_book groups] objectAtIndex: groupIndex];
  [g addMember: p];
}

- (BOOL) applicationShouldTerminate: (NSApplication*) app
{
  if([_book hasUnsavedChanges])
    {
      int retval =
	NSRunAlertPanel(_(@"Save Changes?"),
			_(@"You have made changes to the database.\n"
			  @"Should these changes be saved?"),
			_(@"Save and Quit"), _(@"Quit without saving"),
			_(@"Don't quit"), nil);
      switch(retval)
	{
	case 1:
	  if(![_book save])
	    {
	      NSRunAlertPanel(_(@"Couldn't save"),
			      _(@"The database could not be saved!"),
			      _(@"OK"), nil, nil, nil);
	      return NO;
	    }
	  return YES;
	case 0:
	  return YES;
	default:
	  return NO;
	}
    }
  else
    return YES;
}
@end
