// ADPersonPropertyView.m (this is -*- ObjC -*-)
// 
// \author: Bjrn Giesler <giesler@ira.uka.de>
// 
// Address Book Application for GNUstep
// 
// $Author: giesler $
// $Locker:  $
// $Revision: 1.3 $
// $Date: 2002/08/19 08:41:28 $

/* system includes */
#include <Foundation/Foundation.h>

/* my includes */
#include "ADPersonPropertyView.h"
#include "ADPersonPropertyView+Private.h"

@implementation NSString(ADPersonPropertySupport)
- (NSString*) stringByAbbreviatingToFitWidth: (int) width
				      inFont: (NSFont*) font
{
  if([font widthOfString: self] <= width) return self;
  NSAssert([self length]>3, @"String too short");

  int index = [self length]-3;
  while(index>=0)
    {
      NSString *str = [[self substringToIndex: index]
			stringByAppendingString: @"..."];
      if([font widthOfString: str] <= width) return str;
      index--;
    }
  return nil;
}

- (NSString*) stringByTrimmingWhitespace
{
  NSCharacterSet *wsp = [NSCharacterSet whitespaceAndNewlineCharacterSet];
  return [self stringByTrimmingCharactersInSet: wsp];
}
  
- (BOOL) isEmptyString
{
  NSString *str = [self stringByTrimmingWhitespace];
  if([str isEqualToString: @""]) return YES;
  return NO;
}
@end

@implementation ADPersonPropertyCell
- (void) dealloc
{
  [_details release];
  [super dealloc];
}
- (void) setRect: (NSRect) r
{
  _r = r;
}
- (NSRect) rect
{
  return _r;
}
- (void) setDetails: (id) details
{
  [_details release];
  _details = [details retain];
}
- (id) details
{
  return _details;
}
- (id) copyWithZone: (NSZone*) zone
{
  ADPersonPropertyCell *cell = [super copyWithZone: zone];
  cell->_details = [_details retain];
  return cell;
}
@end


@implementation ADPersonPropertyView
- initWithFrame: (NSRect) frame
{
  _maxLabelWidth = 110;
  _fontSize = 11;
  _font = [[NSFont systemFontOfSize: _fontSize] retain];
  _fontSetExternally = NO;
  _editable = NO;
  _editingCellIndex = -1;

  _clickSel = @selector(clickedOnProperty:withValue:inView:);
  _changeSel = @selector(valueForProperty:changedToValue:inView:);
  _canPerformSel = @selector(canPerformClickForProperty:);
  _widthSel = @selector(view:changedWidthFrom:to:);
  _editInNextSel = @selector(beginEditingInNextViewWithTextMovement:);

  return [super initWithFrame: frame];
}

- (void) dealloc
{
  if(_editingCellIndex || _textObject)
    [self endEditing];
  [_cells release];
  [_person release];
  [_font release];
  [_delegate release];
}

- (void) setDelegate: (id) delegate
{
  [_delegate release];
  _delegate = [delegate retain];
}

- (id) delegate
{
  return _delegate;
}

- (void) setPerson: (ABPerson*) person
{
  [_person release];
  _person = [person retain];
  if(_property) [self layout];
}

- (ABPerson*) person
{
  return _person;
}

- (void) setProperty: (NSString*) property
{
  _property = [property copy];
  if(_person) [self layout];
}

- (NSString*) property
{
  return _property;
}

- (BOOL) updatePersonWithMultiValueFromCell: (ADPersonPropertyCell*) cell
{
  NSString *label, *identifier; id value; int i;
  ADPropertyType type; ADMutableMultiValue *mv;

  identifier = [[cell details] objectForKey: @"Identifier"];
  label = [[cell details] objectForKey: @"Label"];
  value = [cell stringValue];
  if([value isEmptyString]) return NO;

  type = [ADPerson typeOfProperty: _property];

  mv = [_person valueForProperty: _property];
  if(!mv)
    {
      NSLog(@"The person has no value for %@\n", _property);
      return NO;
    }
  
  switch(type)
    {
    case ADMultiDateProperty:
      value = [NSDate dateWithString: value];
      if(!value) return NO;
    case ADMultiStringProperty:
      if(identifier)
	{
	  i = [mv indexForIdentifier: identifier];
	  if(i == NSNotFound)
	    {
	      NSLog(@"Multivalue for %@ has no index for ID %@\n",
		    _property, identifier);
	      return NO;
	    }
	  if(label && ![mv replaceLabelAtIndex: i withLabel: label])
	    return NO;
	  return [mv replaceValueAtIndex: i withValue: value];
	}
      else
	{
	  if(!label)
	    {
	      NSLog(@"No label given for new value %@ of property %@\n",
		    value, _property);
	      return NO;
	    }
	  identifier = [mv addValue: value withLabel: label];
	  if(!identifier) return NO;
	  NSLog(@"New identifier %@ for new value %@ of property %@\n",
		identifier, value, _property);
	  return YES;
	}
      break;
    default:
      NSLog(@"Can't update values of type 0x%x yet.\n", type);
    }

  return NO;
}  

- (BOOL) updatePersonWithValueFromCell: (ADPersonPropertyCell*) cell
{
  ADPropertyType type = [ADPerson typeOfProperty: _property];
  if(type & ADMultiValueMask)
    return [self updatePersonWithMultiValueFromCell: cell];

  id value = [cell stringValue];
  
  switch(type)
    {
    case ADDateProperty:
      value  = [NSDate dateWithString: value];
      if(!value) return NO;
    case ADStringProperty:
      return [_person setValue: value forProperty: _property];
    default:
      NSLog(@"Can't handle type %d yet\n", type);
      return NO;
    }
  return NO;
}

- (void) setDisplaysLabel: (BOOL) yesno
{
  if(_displaysLabel == yesno) return;
  _displaysLabel = yesno;
  if([_cells count]) [self layout];
}
  
- (BOOL) displaysLabel
{
  return _displaysLabel;
}

- (void) setMaxLabelWidth: (int) width
{
  _maxLabelWidth = width;
  if([_cells count]) [self layout];
}

- (int) maxLabelWidth
{
  return _maxLabelWidth;
}

- (int) neededLabelWidth
{
  return _neededLabelWidth;
}

- (void) drawRect: (NSRect) rect
{
  [super drawRect: rect];

  [self lockFocus];

  NSEnumerator *e = [_cells objectEnumerator];
  ADPersonPropertyCell *c;
  while((c = [e nextObject]))
    [c drawWithFrame: [c rect] inView: self];

  [self unlockFocus];
}

- (BOOL) isFlipped
{
  return YES; 
}

- (NSFont*) font
{
  return _font;
}
- (NSFont*) boldFont
{
  return [NSFont boldSystemFontOfSize: [self fontSize]];
}
- (int) fontSize
{
  return _fontSize;
}
- (void) setFontSize: (int) size
{
  _fontSize = size;
  if(!_fontSetExternally)
    {
      [_font release];
      _font = [[NSFont systemFontOfSize: _fontSize] retain];
    }
  if([_cells count]) [self layout];
}

- (void) setFont: (NSFont*) font
{
  [_font release];
  _font = [font retain];
  _fontSetExternally = YES;
  if([_cells count]) [self layout];
}

- (void) setEditable: (BOOL) editable
{
  if(_editable == editable)
    return;
  _editable = editable;
  _editingCellIndex = -1;
  [self layout];
}
- (BOOL) isEditable
{
  return _editable;
}

/*
 * editing
 */

- (void) endEditing
{
  if(_editingCellIndex != -1)
    {
      id cell = [_cells objectAtIndex: _editingCellIndex];
      [cell endEditing: _textObject];
      [_textObject removeFromSuperview];
      _textObject = nil;
      _editingCellIndex = -1;
    }

  if(_textObject)
    {
      [_textObject resignFirstResponder];
      [_window makeFirstResponder: _window];
    }
}

- (void) beginEditingInCellAtIndex: (int) i
		    becauseOfEvent: (NSEvent*) e
{
  [self endEditing];

  ADPersonPropertyCell *c;
  NS_DURING {
    c =  [_cells objectAtIndex: i];
  } NS_HANDLER {
    return;
  } NS_ENDHANDLER;
  
  if(![c isEditable]) return;
  _editingCellIndex = i;
  NSRect r = [c rect];
      
  NSText *t = [_window fieldEditor: YES forObject: c];
  _textObject = [c setUpFieldEditorAttributes: t];
  [_textObject setBackgroundColor: [NSColor orangeColor]];
  [_textObject setTextColor: [NSColor blackColor]];
  [_textObject setDrawsBackground: YES];
  [_textObject setString: [c stringValue]];
  
  if([[c stringValue] hasPrefix: @"["])
    {
      [c setStringValue: @""];
      r.size.width = [[c font] widthOfString: @""];
    }
      
  r.size.width += 4; // make the cursor fit too

  if(e)
    [c editWithFrame: r
       inView: self
       editor: _textObject
       delegate: self
       event: e];
  else
    {
      // HACK: We must create our own event here, since we can't
      // pass nil as event argument to
      // editWithFrame:inView:...
      // REASON: In that method, the event is asked for its event
      // type, which (for a nil event) is 0==NSLeftMouseDown. The
      // effect is that the NSCell sits there waiting for the
      // mouse up :-)
      // REMEDY: Maybe fix this by checking for nil event in
      // NSCell editWithFrame:inView:...?
      e = [NSEvent keyEventWithType: NSKeyDown
		   location: NSMakePoint(0, 0)
		   modifierFlags: 0
		   timestamp: 0
		   windowNumber: 0
		   context: nil
		   characters: @"\t"
		   charactersIgnoringModifiers: @"\t"
		   isARepeat: NO
		   keyCode: '\t'];
      [c editWithFrame: r
	 inView: self
	 editor: _textObject
	 delegate: self
	 event: e];
      [_textObject setSelectedRange: NSMakeRange(0, [[c stringValue]
						      length])];
    }

  [self setNeedsDisplay: YES];
}

- (BOOL) hasEditableCells
{
  int i;
  for(i=0; i<[_cells count]; i++)
    if([[_cells objectAtIndex: i] isEditable])
      return YES;
  return NO;
}
  
- (void) beginEditingInCellAtIndex: (int) i
{
  while(![[_cells objectAtIndex: i] isEditable])
    i++;
  [self beginEditingInCellAtIndex: i becauseOfEvent: nil];
}

/*
 * action methods
 */

- (void) mouseDown: (NSEvent*) event
{
  NSPoint p = [self convertPoint: [event locationInWindow]
		    fromView: nil];

  if(_editable)
    if(_delegate)
      [_delegate viewWillBeginEditing: self];
  
  int i; ADPersonPropertyCell *c = nil;
  for(i=0; i<[_cells count]; i++)
    {
      c = [_cells objectAtIndex: i];
      if(NSPointInRect(p, [c rect]))
	break;
    }
  if(!c) return;

  if(_editable)
    [self beginEditingInCellAtIndex: i becauseOfEvent: event];
  else if(_delegate)
    if([_delegate respondsToSelector: _clickSel])
      [_delegate clickedOnProperty: _property
		 withValue: [c stringValue]
		 inView: self];
}


- (void) textDidEndEditing: (NSNotification*) aNotification
{
  int textMovement = -1;
  int cellIndex;
  id c;

  c = [_cells objectAtIndex: _editingCellIndex];
  [c endEditing: _textObject];
  [c setStringValue: [[[_textObject string] copy] autorelease]];

  [self updatePersonWithValueFromCell: c];
  
  NSDictionary *dict = [aNotification userInfo];
  if(dict)
    {
      id num = [dict objectForKey: @"NSTextMovement"];
      if(num) textMovement = [num intValue];
    }

  cellIndex = _editingCellIndex;
  _editingCellIndex = -1;
  [_textObject removeFromSuperview];
  _textObject = nil;

  if(_delegate)
    if([_delegate respondsToSelector: _changeSel])
      [_delegate valueForProperty: _property
		 changedToValue: [c stringValue]
		 inView: self];

  // Ended editing by any special text movement?
  switch(textMovement)
    {
    case NSTabTextMovement:
    case NSReturnTextMovement:
    case NSBacktabTextMovement:
      if(_delegate)
	if([_delegate respondsToSelector: _editInNextSel])
	  [_delegate beginEditingInNextViewWithTextMovement: textMovement];
      break;
    default:
      break;
    }
  
  [self setNeedsDisplay: YES];
}

- (void) textDidChange: (NSNotification*) aNotification
{
  id c = [_cells objectAtIndex: _editingCellIndex];

  NSPoint o = [c rect].origin;
  NSSize s = [c rect].size;
  s.width = o.x + [[c font] widthOfString: [_textObject string]] + 4;

  NSSize oldSize = [self frame].size;

  s.width = MAX(s.width, _requiredSize.width);
  s.height = MAX(s.height, _requiredSize.height);
  
  [self setFrameSize: s];

  [[_textObject superview] setFrameSize: s];
  [_textObject setFrameSize: s];
  [_textObject setNeedsDisplay: YES];
  [self setNeedsDisplay: YES];

  if(_delegate && [_delegate respondsToSelector: _widthSel])
    [_delegate view: self changedWidthFrom: oldSize.width to: s.width];
}

@end

