// ADMultiValue.m (this is -*- ObjC -*-)
// 
// \author: Bjrn Giesler <giesler@ira.uka.de>
// 
// Address Book API for GNUstep
// 
// $Author: giesler $
// $Locker:  $
// $Revision: 1.3 $
// $Date: 2002/08/19 08:41:28 $

/* system includes */
/* (none) */

/* my includes */
#include "ADMultiValue.h"

#define IS_A(obj,cls) [obj isKindOf: [cls class]]

static ADPropertyType _propTypeFromDict(NSDictionary *dict)
{
  id obj = [dict objectForKey: @"Value"];
  
  if(IS_A(obj, NSString))
    return ADStringProperty;
  if(IS_A(obj, NSDate))
    return ADDateProperty;
  if(IS_A(obj, NSArray))
    return ADArrayProperty;
  if(IS_A(obj, NSDictionary))
    return ADDictionaryProperty;
  if(IS_A(obj, NSData))
    return ADDataProperty;
  if(IS_A(obj, NSValue))
    return ADIntegerProperty;

  return ADErrorInProperty;
}

@interface ADMultiValue (Private)
- initWithMultiValue: (ADMultiValue*) mv;
- (NSArray*) array;
- (ADPropertyType) type;
@end

@interface ADMutableMultiValue (Private)
- initWithMultiValue: (ADMultiValue*) mv;
- postChange;
@end

@implementation ADMultiValue
- initWithType: (ADPropertyType) type
{
  _arr = [[NSMutableArray alloc] initWithCapacity: 5];
  _primaryId = nil;
  _type = type;
  return [super init];
}

- (void) dealloc
{
  [_arr release];
  [_primaryId release];
}

- (unsigned int) count
{
  return [_arr count];
}

- (id) valueAtIndex: (int) index
{
  return [[_arr objectAtIndex: index] objectForKey: @"Value"];
}

- (NSString*) labelAtIndex: (int) index
{
  return [[_arr objectAtIndex: index] objectForKey: @"Label"];
}

- (NSString*) identifierAtIndex: (int) index
{
  return [[_arr objectAtIndex: index] objectForKey: @"ID"];
}

- (int) indexForIdentifier: (NSString*) identifier
{
  int i;
  for(i=0; i<[_arr count]; i++)
    if([[[_arr objectAtIndex: i] objectForKey: @"ID"]
	 isEqualToString: identifier])
      return i;
  return NSNotFound;
}

- (NSString*) primaryIdentifier
{
  return _primaryId;
}

- (ADPropertyType) propertyType
{
  if(![_arr count])
    return ADErrorInProperty;

  NSEnumerator *e = [_arr objectEnumerator];
  id obj = [e nextObject];
  ADPropertyType assumedType = _propTypeFromDict(obj);
  while((obj = [e nextObject]))
    if(assumedType != _propTypeFromDict(obj))
      return ADErrorInProperty;

  return assumedType;
}

- (NSString*) description
{
  return [_arr description];
}

- (id) copyWithZone: (NSZone*) zone
{
  return [[ADMultiValue alloc] initWithMultiValue: self];
}

- (id) mutableCopyWithZone: (NSZone*) zone
{
  return [[ADMutableMultiValue alloc] initWithMultiValue: self];
}
@end

@implementation ADMultiValue (Private)
- initWithMultiValue: (ADMultiValue*) mv
{
  [super init];
  _arr = [[mv array] mutableCopy];
  _primaryId = [[mv primaryIdentifier] copy];
  _type = [mv type];
  return self;
}

- (NSArray*) array
{
  return _arr;
}

- (ADPropertyType) type
{
  return _type;
}
@end

#define POSTCHANGE [[NSNotificationCenter defaultCenter] postNotificationName: @"_MVChg" object: self];

@implementation ADMutableMultiValue
- initWithType: (ADPropertyType) type
{
  _nextId = 0;
  return [super initWithType: type];
}

- (NSString*) _nextValidID
{
  NSEnumerator *e = [_arr objectEnumerator];
  NSDictionary *dict;
  int max = 0;

  while((dict = [e nextObject]))
    max = MAX(max, [[dict objectForKey: @"ID"] intValue]);
  
  return [NSString stringWithFormat: @"%d", max+1];
}

- (NSString*) addValue: (id) value
	     withLabel: (NSString*) label
{
  NSString *identifier = [self _nextValidID];
  NSMutableDictionary *dict =
    [NSMutableDictionary dictionaryWithObjectsAndKeys:
			   value, @"Value",
			 label, @"Label",
			 identifier, @"ID",
			 nil];

  [_arr addObject: dict];
  POSTCHANGE;
  return identifier;
}

- (NSString *) insertValue: (id) value
		 withLabel: (NSString*) label
		   atIndex: (int) index
{
  NSString* identifier = [self _nextValidID];
  NSMutableDictionary *dict =
    [NSMutableDictionary dictionaryWithObjectsAndKeys:
			   value, @"Value",
			 label, @"",
			 identifier, @"ID",
			 nil];
  [_arr insertObject: dict atIndex: index];
  POSTCHANGE;
  return identifier;
}

- (BOOL) removeValueAndLabelAtIndex: (int) index
{
  if(index < 0 || index >= [_arr count]) return NO;
  [_arr removeObjectAtIndex: index];
  POSTCHANGE;
  return YES;
}

- (BOOL) replaceValueAtIndex: (int) index
		   withValue: (id) value
{
  if(index < 0 || index >= [_arr count]) return NO;
  NSMutableDictionary *dict = [_arr objectAtIndex: index];
  [dict setObject: value forKey: @"Value"];
  POSTCHANGE;
  return YES;
}

- (BOOL) replaceLabelAtIndex: (int) index
		   withLabel: (NSString*) label
{
  if(index < 0 || index >= [_arr count]) return NO;
  NSMutableDictionary *dict = [_arr objectAtIndex: index];
  [dict setObject: label forKey: @"Label"];
  POSTCHANGE;
  return YES;
}

- (BOOL)setPrimaryIdentifier:(NSString *)identifier
{
  [_primaryId release];
  _primaryId = [identifier retain];
  POSTCHANGE;
  return YES;
}
@end

@implementation ADMutableMultiValue(AddressesExtensions)
- (BOOL) addValue: (id) value
	withLabel: (NSString*) label
       identifier: (NSString*) identifier
{
  if([self indexForIdentifier: identifier] != NSNotFound)
    return NO;
  
  NSMutableDictionary *dict =
    [NSMutableDictionary dictionaryWithObjectsAndKeys:
			   value, @"Value",
			 label, @"Label",
			 identifier, @"ID",
			 nil];

  [_arr addObject: dict];
  POSTCHANGE;
  return YES;
}
@end
